from __future__ import annotations

import logging

from django.conf import settings

pytest_plugins = []  # type: ignore


# Settings fixtures to bootstrap our tests
def pytest_configure(config):
    logging.disable(logging.CRITICAL)

    settings.configure(
        ALLOWED_HOSTS=["*"],
        CACHES={
            "default": {
                "BACKEND": "django.core.cache.backends.dummy.DummyCache",
            }
        },
        DATABASES={
            "default": {
                "ENGINE": "django.db.backends.sqlite3",
                "NAME": ":memory:",
            },
        },
        EMAIL_BACKEND="django.core.mail.backends.locmem.EmailBackend",
        INSTALLED_APPS=[
            "django.contrib.contenttypes",
            "django_twc_toolbox",
            "tests.dummy",
        ],
        LOGGING_CONFIG=None,
        PASSWORD_HASHERS=["django.contrib.auth.hashers.MD5PasswordHasher"],
        SECRET_KEY="NOTASECRET",
        USE_TZ=True,
    )



from __future__ import annotations

from django.db import models


class DateOrderableModel(models.Model):
    date = models.DateField()


class DateTimeOrderableModel(models.Model):
    date = models.DateTimeField()



from __future__ import annotations

import datetime
import itertools
import math
import random
from dataclasses import asdict
from dataclasses import dataclass

import pytest
from django.core.paginator import EmptyPage
from django.core.paginator import Page
from django.core.paginator import PageNotAnInteger
from django.db.models.query import QuerySet
from django.utils import timezone
from model_bakery import baker

from django_twc_toolbox.paginator import DatePaginator

from .dummy.models import DateOrderableModel
from .dummy.models import DateTimeOrderableModel


@dataclass
class ModelClassParams:
    """
    Stores parameters for a pytest parameterized test case.

    Intended to be used as a parameter to a parameterized test case along
    with the indirect `model_data_queryset` fixture.
    """

    model_class: type[DateOrderableModel | DateTimeOrderableModel]
    number_of_days: int

    def __iter__(self):
        return iter(asdict(self).values())

    @classmethod
    def from_request(cls, request):
        return cls(
            model_class=request.param.model_class,
            number_of_days=request.param.number_of_days,
        )


@pytest.fixture
def model_data_queryset(request, db):
    """
    Sets up a queryset of the specified model class with the specified number of days of data
    for use in pagination tests.

    Depending on the test parameterization, it uses `ModelClassParams` to determine the model class
    and the number of days for data generation. If not parameterized with `ModelClassParams`,
    it defaults to `DateOrderableModel` with 90 days of data.

    The fixture supports creating data for both `DateOrderableModel` and `DateTimeOrderableModel`.
    For `DateTimeOrderableModel`, it generates two data points per day.

    The DatePaginator expects an ordered data structure, so the data is returned ordered to make
    all the tests do not need to worry about the initial ordering. In the tests that need to verify
    the ordering, they can apply their own sorting to the data.
    """
    # Check if the request has the expected parameterization
    if hasattr(request, "param") and isinstance(request.param, ModelClassParams):
        model_class, number_of_days = ModelClassParams.from_request(request)
    else:
        # Default values for tests that don't use ModelClassParams
        model_class = DateOrderableModel  # or any default model
        number_of_days = 90  # default number of days

    if model_class == DateOrderableModel:
        date_values = (
            timezone.now() - datetime.timedelta(days=i) for i in range(number_of_days)
        )
    else:
        current_time = timezone.now()
        date_values = itertools.chain.from_iterable(
            (
                current_time - datetime.timedelta(days=i, hours=12),
                current_time - datetime.timedelta(days=i),
            )
            for i in range(number_of_days)
        )

    baker.make(
        model_class,
        date=itertools.cycle(date_values),
        _quantity=number_of_days,
    )

    return model_class.objects.all().order_by("date")


@pytest.fixture(params=["queryset", "list", "tuple"])
def objects(request, model_data_queryset):
    """
    Provides the model data needed in different formats to make sure our custom
    DatePaginator can handle all the data structures a Django Paginator should handle.

    It uses the `model_data_queryset` fixture to get the initial queryset and then
    provides it in three different structures: QuerySet, list, and tuple.
    """
    if request.param == "queryset":
        ret = model_data_queryset
    elif request.param == "tuple":
        ret = tuple(model_data_queryset)
    else:
        ret = list(model_data_queryset)

    return ret


class TestDatePaginator:
    @pytest.mark.parametrize(
        "model_data_queryset,days_per_page,expected_num_pages",
        [
            # date
            # create 90 days of data, paginate by 30 days, expect 3 pages
            (
                ModelClassParams(model_class=DateOrderableModel, number_of_days=90),
                30,
                3,
            ),
            # datetime
            # create 90 days of data (2 per day), paginate by 30 days, expect 2 pages
            (
                ModelClassParams(model_class=DateTimeOrderableModel, number_of_days=90),
                30,
                2,
            ),
            # date
            # create 180 days of data, paginate by 30 days, expect 6 pages
            (
                ModelClassParams(model_class=DateOrderableModel, number_of_days=180),
                30,
                6,
            ),
            # datetime
            # create 180 days of data (2 per day), paginate by 30 days, expect 3 pages
            (
                ModelClassParams(
                    model_class=DateTimeOrderableModel, number_of_days=180
                ),
                30,
                3,
            ),
            # date
            # create 90 days of data, paginate by 60 days, expect 2 pages
            (
                ModelClassParams(model_class=DateOrderableModel, number_of_days=90),
                60,
                2,
            ),
            # date
            # create 90 days of data (2 per day), paginate by 60 days, expect 1 page
            (
                ModelClassParams(model_class=DateTimeOrderableModel, number_of_days=90),
                60,
                1,
            ),
        ],
        indirect=["model_data_queryset"],
    )
    def test_num_pages(self, objects, days_per_page, expected_num_pages):
        paginator = DatePaginator(
            objects, "date", datetime.timedelta(days=days_per_page)
        )

        assert paginator.num_pages == expected_num_pages

    def test_first_page(self, objects):
        paginator = DatePaginator(objects, "date", datetime.timedelta(days=30))
        first_page = paginator.page(1)

        first_segment_start, first_segment_end = paginator.date_segments[0]

        # Count objects in the first segment based on the type of 'objects'
        if isinstance(objects, QuerySet):  # type: ignore[misc]
            objects_in_first_segment = objects.filter(
                date__gte=first_segment_start,
                date__lt=first_segment_end,
            ).count()
        else:
            objects_in_first_segment = sum(
                1
                for entry in objects
                if first_segment_start <= entry.date < first_segment_end
            )

        assert len(first_page.object_list) == objects_in_first_segment

    def test_page_out_of_range(self, objects):
        paginator = DatePaginator(objects, "date", datetime.timedelta(days=30))

        with pytest.raises(EmptyPage):
            paginator.page(4)  # Only 3 pages exist

    def test_with_empty_queryset(self):
        empty = DateOrderableModel.objects.none().order_by("date")
        paginator = DatePaginator(empty, "date", datetime.timedelta(days=30))

        assert paginator.count == 0
        assert paginator.num_pages == 0

    def test_with_empty_list(self):
        paginator = DatePaginator([], "date", datetime.timedelta(days=30))

        assert paginator.count == 0
        assert paginator.num_pages == 0

    def test_with_single_entry(self, db):
        baker.make("dummy.DateOrderableModel", date=timezone.now().date())
        obj = DateOrderableModel.objects.all().order_by("-date")
        paginator = DatePaginator(obj, "date", datetime.timedelta(days=30))

        assert paginator.num_pages == 1
        assert len(paginator.page(1).object_list) == 1

    def test_large_date_range(self, objects):
        large_range = datetime.timedelta(days=365)
        paginator = DatePaginator(objects, "date", large_range)

        assert paginator.num_pages == 1
        assert len(paginator.page(1).object_list) == len(objects)

    def test_small_date_range(self, objects):
        small_range = datetime.timedelta(days=1)
        paginator = DatePaginator(objects, "date", small_range)

        assert paginator.num_pages == len(objects)

    def test_date_range_boundary_cases(self, db, objects):
        if isinstance(objects, QuerySet):  # type: ignore[misc]
            earliest_date = objects.earliest("date").date
        else:
            earliest_date = min(entry.date for entry in objects)

        paginator = DatePaginator(objects, "date", datetime.timedelta(days=30))
        start_of_first_segment, _ = paginator.date_segments[0]

        # Ensure the boundary date falls within the first segment
        boundary_date = max(earliest_date, start_of_first_segment)

        # Create new objects at the boundary date
        new_objects = baker.make(
            "dummy.DateOrderableModel", date=boundary_date, _quantity=5
        )

        # If 'objects' is a tuple/list, manually add these new objects to it
        if not isinstance(objects, QuerySet):  # type: ignore[misc]
            if isinstance(objects, tuple):
                objects = objects + tuple(new_objects)
            else:
                objects.extend(new_objects)

        first_page = paginator.page(1)

        assert all(obj.date >= boundary_date for obj in first_page.object_list)

    def test_non_sequence_page_number(self, objects):
        if isinstance(objects, QuerySet):  # type: ignore[misc]
            random_objects = objects.order_by("?")
        else:
            random_objects = list(objects)
            random.shuffle(random_objects)
            if isinstance(objects, tuple):
                random_objects = tuple(random_objects)

        with pytest.raises(ValueError):
            DatePaginator(random_objects, "date", datetime.timedelta(days=30))

    @pytest.mark.parametrize(
        "model_data_queryset",
        [
            ModelClassParams(model_class=DateOrderableModel, number_of_days=90),
            ModelClassParams(model_class=DateTimeOrderableModel, number_of_days=180),
        ],
        indirect=["model_data_queryset"],
    )
    def test_chronological_ordering(self, objects):
        paginator = DatePaginator(objects, "date", datetime.timedelta(days=30))

        first_page_item_date = paginator.page(1).object_list[0].date
        middle_page_item_date = paginator.page(2).object_list[0].date
        last_page_item_date = paginator.page(3).object_list[0].date

        assert first_page_item_date < middle_page_item_date
        assert middle_page_item_date < last_page_item_date

        for item in paginator.page(1).object_list:
            assert item.date >= first_page_item_date

    @pytest.mark.parametrize(
        "model_data_queryset",
        [
            ModelClassParams(model_class=DateOrderableModel, number_of_days=90),
            ModelClassParams(model_class=DateTimeOrderableModel, number_of_days=180),
        ],
        indirect=["model_data_queryset"],
    )
    def test_reversed_ordering(self, objects):
        if isinstance(objects, QuerySet):  # type: ignore[misc]
            objects = objects.order_by("-date")
        else:
            objects = list(reversed(objects))

        paginator = DatePaginator(objects, "date", datetime.timedelta(days=30))

        first_page_item_date = paginator.page(1).object_list[0].date
        middle_page_item_date = paginator.page(2).object_list[0].date
        last_page_item_date = paginator.page(3).object_list[0].date

        assert first_page_item_date > middle_page_item_date
        assert middle_page_item_date > last_page_item_date

        previous_item = first_page_item_date
        for item in paginator.page(1).object_list:
            assert item.date <= first_page_item_date
            assert item.date <= previous_item
            previous_item = item.date

    def test_paginator_init(self, objects):
        paginator = DatePaginator(
            objects,
            "date",
            datetime.timedelta(days=10),
            orphans=5,
            allow_empty_first_page=False,
        )

        assert paginator.object_list == objects
        assert paginator.per_page == 1  # As it paginates by date
        assert paginator.orphans == 5
        assert paginator.allow_empty_first_page is False

    def test_paginator_get_page(self, objects):
        paginator = DatePaginator(objects, "date", datetime.timedelta(days=10))

        # Valid page number
        page = paginator.get_page(1)

        assert isinstance(page, Page)
        assert page.number == 1

        # Invalid page number
        invalid_page = paginator.get_page(1000)

        assert invalid_page.number == paginator.num_pages  # Should return the last page

    def test_paginator_page_function(self, objects):
        paginator = DatePaginator(objects, "date", datetime.timedelta(days=10))

        page = paginator.page(1)

        assert isinstance(page, Page)
        assert page.number == 1

        with pytest.raises(PageNotAnInteger):
            paginator.page("not_a_number")

        with pytest.raises(EmptyPage):
            paginator.page(0)  # Page number less than 1

        with pytest.raises(EmptyPage):
            paginator.page(paginator.num_pages + 1)  # Page number out of range

    def test_paginator_get_elided_page_range(self, objects):
        paginator = DatePaginator(objects, "date", datetime.timedelta(days=10))

        elided_range = paginator.get_elided_page_range(number=1)
        elided_range_list = list(elided_range)  # Convert generator to list

        # The list should contain integers representing page numbers and possibly the ELLIPSIS
        assert all(
            isinstance(item, int) or item == paginator.ELLIPSIS
            for item in elided_range_list
        )

    @pytest.mark.parametrize(
        "model_data_queryset",
        [
            ModelClassParams(model_class=DateOrderableModel, number_of_days=90),
            ModelClassParams(model_class=DateTimeOrderableModel, number_of_days=180),
        ],
        indirect=["model_data_queryset"],
    )
    def test_paginator_attributes(self, objects):
        date_range = datetime.timedelta(days=10)
        paginator = DatePaginator(objects, "date", date_range)

        # Check ELLIPSIS
        assert paginator.ELLIPSIS == "…"

        # Check count
        assert paginator.count == len(objects)

        # Check num_pages
        if isinstance(objects, QuerySet):  # type: ignore[misc]
            earliest_date = objects.earliest("date").date
            latest_date = objects.latest("date").date
        else:
            dates = [entry.date for entry in objects]
            earliest_date = min(dates)
            latest_date = max(dates)

        date_span = (latest_date - earliest_date).days + 1  # +1 to include the last day
        expected_num_pages = math.ceil(date_span / date_range.days)

        assert paginator.num_pages == expected_num_pages

        # Check page_range
        assert isinstance(paginator.page_range, range)
        assert len(paginator.page_range) == paginator.num_pages


class TestDatePage:
    def test_page_start_date(self, objects):
        paginator = DatePaginator(objects, "date", datetime.timedelta(days=10))
        page = paginator.page(1)

        start_date, _ = paginator.date_segments[0]

        assert page.start_date == start_date

    def test_page_end_date(self, objects):
        paginator = DatePaginator(objects, "date", datetime.timedelta(days=10))
        page = paginator.page(1)

        _, end_date = paginator.date_segments[0]

        assert page.end_date == end_date

    def test_page_min_date(self, objects):
        paginator = DatePaginator(objects, "date", datetime.timedelta(days=10))
        page = paginator.page(1)

        start_date, _ = paginator.date_segments[0]

        assert page.min_date == start_date

    def test_page_max_date(self, objects):
        paginator = DatePaginator(objects, "date", datetime.timedelta(days=10))
        page = paginator.page(1)

        assert page.max_date.date() == timezone.now().date()

        page = paginator.page(2)

        _, end_date = paginator.date_segments[1]
        assert page.max_date == end_date

    def test_page_date_range(self, objects):
        paginator = DatePaginator(objects, "date", datetime.timedelta(days=10))
        page = paginator.page(1)

        start_date, _ = paginator.date_segments[0]

        assert page.date_range == (start_date, page.max_date)

    def test_page_has_next(self, objects):
        paginator = DatePaginator(objects, "date", datetime.timedelta(days=10))
        page = paginator.page(1)

        has_next = page.number < paginator.num_pages

        assert page.has_next() == has_next

    def test_page_has_previous(self, objects):
        paginator = DatePaginator(objects, "date", datetime.timedelta(days=10))
        page = paginator.page(1)

        has_previous = page.number > 1

        assert page.has_previous() == has_previous

    def test_page_has_other_pages(self, objects):
        paginator = DatePaginator(objects, "date", datetime.timedelta(days=10))
        page = paginator.page(1)

        has_other_pages = paginator.num_pages > 1

        assert page.has_other_pages() == has_other_pages

    def test_page_next_page_number(self, objects):
        paginator = DatePaginator(objects, "date", datetime.timedelta(days=10))
        page = paginator.page(1)

        assert page.next_page_number() == page.number + 1

        page = paginator.page(paginator.num_pages)  # Last page

        with pytest.raises(EmptyPage):
            page.next_page_number()

    def test_page_previous_page_number(self, objects):
        paginator = DatePaginator(objects, "date", datetime.timedelta(days=10))
        page = paginator.page(paginator.num_pages)  # Last page

        assert page.previous_page_number() == page.number - 1

        page = paginator.page(1)

        with pytest.raises(EmptyPage):
            page.previous_page_number()

    def test_page_start_index(self, objects):
        paginator = DatePaginator(objects, "date", datetime.timedelta(days=10))
        page = paginator.page(1)

        assert page.start_index() == (page.number - 1) * page.paginator.per_page + 1

    def test_page_end_index(self, objects):
        paginator = DatePaginator(objects, "date", datetime.timedelta(days=10))
        page = paginator.page(1)

        assert page.end_index() == min(
            page.number * page.paginator.per_page, page.paginator.count
        )

    def test_page_attributes(self, objects):
        paginator = DatePaginator(objects, "date", datetime.timedelta(days=10))
        page = paginator.page(1)

        assert hasattr(page.object_list, "__len__") or hasattr(
            page.object_list, "count"
        )
        assert page.number == 1
        assert page.paginator == paginator



from __future__ import annotations

import datetime
from typing import TYPE_CHECKING

from django.core.paginator import Page
from django.core.paginator import Paginator
from django.db.models.query import QuerySet
from django.utils import timezone
from django.utils.functional import cached_property

if TYPE_CHECKING:
    from typing import Any

    from django.core.paginator import _SupportsPagination


class DatePaginator(Paginator):
    def __init__(
        self,
        object_list: _SupportsPagination,
        date_field: str,
        date_range: datetime.timedelta,
        **kwargs,
    ) -> None:
        self.date_field = date_field
        self.date_range = date_range
        super().__init__(
            object_list,
            1,  # per_page is 1 as we paginate by date
            **kwargs,
        )

    @cached_property
    def date_segments(self) -> list[tuple[datetime.datetime, datetime.datetime]]:
        # Check if the object_list is empty.
        # Check for `exists()` first for performance, falling back in case it's
        # not a QuerySet.
        if isinstance(self.object_list, QuerySet):  # type: ignore[misc]
            if not self.object_list.exists():
                return []
        elif not self.object_list:
            return []

        if isinstance(self.object_list, QuerySet):  # type: ignore[misc]
            first_obj = self.object_list.first()
            last_obj = self.object_list.last()
        else:
            first_obj = self.object_list[0]
            last_obj = self.object_list[-1]

        first_date = getattr(first_obj, self.date_field)
        last_date = getattr(last_obj, self.date_field)

        segments = []
        current_start_date = first_date
        if self._is_chronological():
            # if chronological, we are moving forward in time through the `object_list`
            # so any time we need to get the next segment's end date, we need to
            # add the date_range. to start we add the date_range to the first date.
            current_end_date = first_date + self.date_range
            # less than or equal because we are moving forwards in time and dates
            # in the future are greater than dates in the past
            # yesterday < today < tomorrow
            while current_end_date <= last_date:
                # keep the tuple ordering consistent with the chronological order
                # so start_date < end_date
                segments.append((current_start_date, current_end_date))
                current_start_date = current_end_date
                current_end_date += self.date_range
            # Append the last segment to cover any remaining dates not included in the
            # previous segments. This is necessary because the date range defined by
            # `date_range` might not perfectly divide the total span of dates. This final
            # segment captures any dates from the end of the last segment up to and
            # including `last_date`. We add one day to `last_date` to ensure the entire
            # day is covered.
            segments.append(
                (current_start_date, last_date + datetime.timedelta(days=1))
            )
        else:
            # if not chronological, it means we are moving backwards in time through
            # the `object_list` so this time we subtract the date_range to get the
            # next segment's end date. to start we subtract the date_range from the
            # first date.
            current_end_date = first_date - self.date_range
            # again, greater than or equal because we are moving backwards in time
            # tomorrow > today > yesterday
            while current_end_date >= last_date:
                # keep the tuple ordering consistent with the reverse chronological
                # order so end_date > start_date
                # segments.append((current_end_date, current_start_date))
                segments.append((current_start_date, current_end_date))
                current_start_date = current_end_date
                current_end_date -= self.date_range
            # subtract because reverse
            # segments.append(
            #     (last_date - datetime.timedelta(days=1), current_start_date)
            # )
            segments.append(
                (current_start_date, last_date - datetime.timedelta(days=1))
            )

        return segments

    def page(self, number: int | str) -> DatePage:
        number = self.validate_number(number)
        start_date, end_date = self.date_segments[number - 1]

        object_list: QuerySet[Any] | list[Any]

        if isinstance(self.object_list, QuerySet):  # type: ignore[misc]
            # For QuerySet, filter based on date range
            if self._is_chronological():
                object_list = self.object_list.filter(
                    **{
                        f"{self.date_field}__gte": start_date,
                        f"{self.date_field}__lt": end_date,
                    }
                )
            else:
                object_list = self.object_list.filter(
                    **{
                        f"{self.date_field}__lte": start_date,
                        f"{self.date_field}__gt": end_date,
                    }
                ).order_by(f"-{self.date_field}")
        else:
            # For non-QuerySet, manually filter and sort
            if self._is_chronological():
                object_list = [
                    obj
                    for obj in self.object_list
                    # yesterday < today < tomorrow
                    if start_date <= getattr(obj, self.date_field) < end_date
                ]
            else:
                object_list = [
                    obj
                    for obj in self.object_list
                    # tomorrow > today > yesterday
                    if start_date >= getattr(obj, self.date_field) > end_date
                ]

            # Apply sorting based on the initial order
            object_list.sort(
                key=lambda obj: getattr(obj, self.date_field),
                reverse=not self._is_chronological(),
            )

        return self._get_page(object_list, number, self, start_date, end_date)

    def _is_chronological(self) -> bool:
        """Check if the object_list is ordered in chronological order

        Chronological
        - oldest to newest
        - e.g. [yesterday, today, tomorrow]
        - yesterday < tomorrow
        - would return True

        Reverse chronological
        - newest to oldest
        - e.g. [tomorrow, today, yesterday]
        - tomorrow > yesterday
        - would return False
        """
        if self.count == 1:
            return True

        if isinstance(self.object_list, QuerySet):  # type: ignore[misc]
            first_obj = self.object_list.first()
            last_obj = self.object_list.last()
        else:
            first_obj = self.object_list[0]
            last_obj = self.object_list[-1]

        first_date = getattr(first_obj, self.date_field)
        last_date = getattr(last_obj, self.date_field)

        return first_date < last_date

    def _get_page(self, *args, **kwargs) -> DatePage:
        return DatePage(*args, **kwargs)

    @cached_property
    def num_pages(self) -> int:
        return len(self.date_segments)

    def _check_object_list_is_ordered(self):
        """Ensure that the object_list is ordered by date_field"""
        if isinstance(self.object_list, QuerySet):  # type: ignore[misc]
            ordering_fields = self.object_list.query.order_by
            if not ordering_fields or not any(
                field in [self.date_field, f"-{self.date_field}"]
                for field in ordering_fields
            ):
                raise ValueError(
                    f"Paginator received an unordered object_list: {self.object_list}. "
                    "DatePaginator only supports ordered object_list instances. "
                    "DatePaginator only supports object_list instances ordered by "
                    f"the specified date_field. Please use .order_by('{self.date_field}') "
                    f"or .order_by('-{self.date_field}') on the queryset."
                )
        else:
            # For lists, check if elements are in ascending or descending order by date_field
            is_ascending = all(
                getattr(x, self.date_field) <= getattr(y, self.date_field)
                for x, y in zip(self.object_list, self.object_list[1:])
            )
            is_descending = all(
                getattr(x, self.date_field) >= getattr(y, self.date_field)
                for x, y in zip(self.object_list, self.object_list[1:])
            )
            if not (is_ascending or is_descending):
                raise ValueError(
                    "Paginator received an unordered list. DatePaginator only supports "
                    f"lists that are ordered by the specified `date_field` - {self.date_field}."
                )


class DatePage(Page):
    def __init__(
        self,
        object_list: _SupportsPagination,
        number: int,
        paginator: DatePaginator,
        start_date: datetime.datetime,
        end_date: datetime.datetime,
    ) -> None:
        super().__init__(object_list, number, paginator)
        self.start_date = start_date
        self.end_date = end_date
        self.min_date = self.start_date
        self.max_date = self.end_date if number != 1 else timezone.now()
        self.date_range = (self.start_date, self.max_date)



from __future__ import annotations

from django.db import models


class TimeStamped(models.Model):
    """
    Abstract model for adding `created_at` and `updated_at` fields.

    This model is based on `TimeStampedModel` from the `django-model-utils`
    package, which can be found at https://github.com/jazzband/django-model-utils.
    License for the original code is below.

    Copyright (c) 2009-2019, Carl Meyer and contributors
    All rights reserved.

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions are
    met:

        * Redistributions of source code must retain the above copyright
          notice, this list of conditions and the following disclaimer.
        * Redistributions in binary form must reproduce the above
          copyright notice, this list of conditions and the following
          disclaimer in the documentation and/or other materials provided
          with the distribution.
        * Neither the name of the author nor the names of other
          contributors may be used to endorse or promote products derived
          from this software without specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
    A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
    OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
    SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
    LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
    OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
    """

    created_at = models.DateTimeField(auto_now_add=True, editable=False)
    updated_at = models.DateTimeField(auto_now=True, editable=False)

    class Meta:
        abstract = True

    def save(self, *args, **kwargs) -> None:
        """
        Overriding the save method in order to make sure that
        modified field is updated even if it is not given as
        a parameter to the update field argument.
        """
        update_fields = kwargs.get("update_fields", None)
        if update_fields:
            kwargs["update_fields"] = set(update_fields).union({"updated_at"})

        super().save(*args, **kwargs)

    def is_edited(self) -> bool:
        return self.created_at != self.updated_at
